#include "model.h"

#include <string>
#include <list>
#include <vector>
#include <iostream>
#include <fstream>

inline
std::list<std::string> split(std::string str, std::string delim)
{
  std::list<std::string> result;

  unsigned cutAt;
  while ((cutAt = str.find_first_of(delim)) != str.npos) {
    if (cutAt > 0) {
      result.push_back(str.substr(0, cutAt));
    }
    else if (cutAt == 0) {
      result.push_back("");
    }
    str = str.substr(cutAt + 1);
  }
  if (str.length() > 0) {
    result.push_back(str);
  }
  return result;
}


inline
glm::vec3 computeNormal(glm::vec3 triangle_pos[3])
{
  glm::vec3 result_normal;
  const glm::vec3 & a = triangle_pos[0];
  const glm::vec3 & b = triangle_pos[1];
  const glm::vec3 & c = triangle_pos[2];
  result_normal = glm::normalize(glm::cross(c - b, a - b));
  return result_normal;
}


Model* loadModelFromObjFile(std::string fileName)
{
  glm::uint numVertices;
  glm::uint numIndices;

  std::ifstream ifs;
  char line[1024];

  std::vector<glm::vec3> positions;
  std::vector<glm::vec3> normals;
  std::vector<glm::vec2> texcoords;
  std::vector<glm::uint> vindices;
  std::vector<glm::uint> nindices;
  std::vector<glm::uint> tindices;

  ifs.open(fileName.c_str(), std::ios::in);
  if (!ifs.is_open()) {
    std::cerr << "Error : Can't file Open \n";
    std::cerr << "File Name : " << fileName << "\n";
    return NULL;
  }

  while (!ifs.eof()) {
    ifs.getline(line, sizeof(line));

    std::list<std::string> l = split(std::string(line), " ");
    if (l.size() < 1) continue;

    std::list<std::string>::iterator it = l.begin();

    std::string head = (*it);
    if (head == "#") {
      //std::cout << "comment" << "\n";
    }
    else if (head == "mtllib" && l.size() >= 2) {
      ++it;
      //std::cout << "mtllib = " << (*it) << "\n";
    }
    else if (head == "g" && l.size() >= 2) {
      ++it;
      //std::cout << "group = " << (*it) << "\n";
    }
    else if (head == "usemtl") {
      ++it;
      //std::cout << "usemtl = " << (*it) << "\n";
    }
    else if (head == "v" && l.size() >= 4) {
      glm::vec3 v;
      ++it; v.x = atof((*it).c_str());
      ++it; v.y = atof((*it).c_str());
      ++it; v.z = atof((*it).c_str());
      positions.push_back( v );
    }
    else if (head == "vn" && l.size() >= 4) {
      glm::vec3 vn;
      ++it; vn.x = atof((*it).c_str());
      ++it; vn.y = atof((*it).c_str());
      ++it; vn.z = atof((*it).c_str());
      normals.push_back( vn );
    }
    else if (head == "vt" && l.size() >= 3) {
      glm::vec2 vt;
      ++it; vt.s = atof((*it).c_str());
      ++it; vt.t = atof((*it).c_str());
      texcoords.push_back( vt );
    }
    else if (head == "f" && l.size() >= 4) {
      int fv[3] = {-1};
      int ft[3] = {-1};
      int fn[3] = {-1};

      int num = l.size()-1;
      for (int i=0; i<num; i++) {
        int idx = i;
        if (idx > 2) idx = 2; // Hint: convert from polygon to triangle

        ++it;
        if (it == l.end()) { break; }
        std::list<std::string> flist = split(std::string(*it), "/");
        std::list<std::string>::iterator fit = flist.begin();
        if (flist.size() == 1) { // v only
          fv[idx] = (*fit == "") ? -1 : atoi((*fit).c_str());
          ft[idx] = -1;
          fn[idx] = -1;
        }
        else if (flist.size() == 2) { // v & t
          fv[idx] = (*fit == "") ? -1 : atoi((*fit).c_str()); ++fit;
          ft[idx] = (*fit == "") ? -1 : atoi((*fit).c_str());
          fn[idx] = -1;
        }
        else if (flist.size() == 3) { // (v & t & n) or (v & 0 & n)
          fv[idx] = (*fit == "") ? -1 : atoi((*fit).c_str()); ++fit;
          ft[idx] = (*fit == "") ? -1 : atoi((*fit).c_str()); ++fit;
          fn[idx] = (*fit == "") ? -1 : atoi((*fit).c_str());
        }

        if (idx >= 2) {
          if ((fv[0] >= 0) && (fv[1] >= 0) && (fv[2] >= 0)) {
            vindices.push_back( fv[0]-1 );
            vindices.push_back( fv[1]-1 );
            vindices.push_back( fv[2]-1 );
          }
          if ((ft[0] >= 0) && (ft[1] >= 0) && (ft[2] >= 0)) {
            tindices.push_back( ft[0]-1 );
            tindices.push_back( ft[1]-1 );
            tindices.push_back( ft[2]-1 );
          }
          if ((fn[0] >= 0) && (fn[1] >= 0) && (fn[2] >= 0)) {
            nindices.push_back( fn[0]-1 );
            nindices.push_back( fn[1]-1 );
            nindices.push_back( fn[2]-1 );
          }
        }
      }
    }
  }

//  std::cout << " vsize = " << positions.size() << "\n";
//  std::cout << " vnsize = " << normals.size() << "\n";
//  std::cout << " vtsize = " << texcoords.size() << "\n";
//  std::cout << " vi = " << vindices.size() << "\n";
//  std::cout << " ni = " << nindices.size() << "\n";
//  std::cout << " ti = " << tindices.size() << "\n";

  ///////////////////////////////////////////////////////////////
  // create Model

  numVertices = positions.size();
  numIndices = vindices.size();

  if (numVertices == 0) { return NULL; }
  if (numIndices == 0) { return NULL; }

  Model* shape = new Model();
  if (!shape) { return NULL; }

  shape->numVertices = numVertices;
  shape->numIndices = numIndices;

  //
  // generate index list
  //

  shape->indices = new glm::uint[numIndices];
  if (!shape->indices) { delete shape; return NULL; }
  for (unsigned i=0; i<numIndices; i++) {
      shape->indices[i] = vindices[i]; // sub 1
  }

  //
  // generate vertex list
  //

  // position
  shape->vertices = new Vertex[numVertices];
  if (!shape->vertices) { delete shape; return NULL; }

  for (unsigned i=0; i<numVertices; i++) {
      shape->vertices[i].position = positions[i];
  }

  // normal
  if (nindices.size() == 0) { // generate normal
    for (unsigned i=0; i<numVertices; i++) {
      int cnt = 0;
      glm::vec3 sum(0.0f,0.0f,0.0f);
      glm::vec3 v[3];
      for (unsigned j=0; j<numIndices/3; j++) {
        if ((i == shape->indices[j*3]) || (i == shape->indices[j*3+1]) || (i == shape->indices[j*3+2])) {
          v[0] = positions[shape->indices[j*3+0]];
          v[1] = positions[shape->indices[j*3+1]];
          v[2] = positions[shape->indices[j*3+2]];
          sum += computeNormal( v );
          cnt++;
        }
      }
      if (cnt > 0)
        shape->vertices[i].normal = glm::normalize(sum / (float)cnt);
    }
  }
  else { // convert normal
    // number of normal == number of position
    for (unsigned i=0; i<numVertices; i++) {
      glm::vec3 sum(0.0f,0.0f,0.0f);
      for (unsigned j=0; j<numIndices; j++) {
        if (i == vindices[j]) {
          sum += normals[nindices[j]];
        }
      }
      shape->vertices[i].normal = glm::normalize(sum);
    }
  }

  // normal
  if (tindices.size() == 0) { // generate texture coordination
    // TODO
  }
  else {
    // TODO
  }

  return shape;
}


//#if 1
//int main(int argc, char* argv[])
//{
//  loadModelFromObjFile("untitled.obj");
//
//
//  return 0;
//}
//#endif
