#ifndef __SHAPE_H_
#define __SHAPE_H_


#include "common.h"
#include "basic_shape.h"


//
// GLSL context
//
struct Context {

  // program object

  GLuint programObj;

  // index

  GLint modelLocation;
  GLint viewLocation;
  GLint projectionLocation;
  GLint texcoordEnLocation;

  GLint positionLocation;
  GLint normalLocation;
  GLint texcoord0Location;
  GLint texture0Location;

  GLint lightPositionLocation;
  GLint lightAmbientLocation;
  GLint lightDiffuseLocation;
  GLint lightSpecularLocation;

  GLint materialAmbientLocation;
  GLint materialDiffuseLocation;
  GLint materialSpecularLocation;
  GLint materialShininessLocation;
};


class Shape
{
protected:
  const Context *context;

  glm::uint textureEnable;
  glm::uint materialEnable;
  GLuint vao[1]; // vertices array buffer
  Model* shape;
  glm::uint numIndices;
  glm::uint numVertices;

  Material  material;

  glm::mat4 modelMatrix;
  glm::vec3 pos;
  glm::vec3 size;

public:
  // constructor
  Shape(const Context* c) : context(c) {
    textureEnable = 0;
    numIndices = 0;
    numVertices = 0;
    pos = glm::vec3(0.0f,0.0f,0.0f);
    size = glm::vec3(1.0f,1.0f,1.0f);
    materialEnable = false;
  }

  void setup(const Model* shape)
  {
    GLuint offset;
    GLuint vbo[2];

    if (!shape) {
      std::cerr << "shape is NULL" << "\n";
    }

    numIndices = shape->numIndices;
    numVertices = shape->numVertices;

    // Create VBO
    glGenBuffers(2, vbo);

    // Create and bind VAO
    glGenVertexArrays(1, vao);
    glBindVertexArray(vao[0]);
    {
      // --- VBO 0 ---
      glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      glBufferData(GL_ARRAY_BUFFER, numVertices*sizeof(Vertex), shape->vertices, GL_STATIC_DRAW);

      offset = 0;
      glVertexAttribPointer(context->positionLocation, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const void*)offset); 
      offset += sizeof(glm::vec3);
      glVertexAttribPointer(context->normalLocation, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const void*)offset); 

      // Enable Attribute
      glEnableVertexAttribArray(context->positionLocation);
      glEnableVertexAttribArray(context->normalLocation);


      // --- VBO 1 ---
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo[1]);
      glBufferData(GL_ELEMENT_ARRAY_BUFFER, numIndices*sizeof(glm::uint), shape->indices, GL_STATIC_DRAW);

    }
    glBindVertexArray(0);

    glDeleteBuffers(2, vbo);

    // material setting
    if (shape->materials.size() > 0) {
      material = shape->materials[0];
      materialEnable = true;
    }
  }

  void render()
  {
    // VAO - vertex, normal, texcoord0
    glBindVertexArray(vao[0]);

    glUniform1i(context->texcoordEnLocation, textureEnable);

    if (materialEnable == 1) {
      glUniform4fv(context->materialAmbientLocation, 1, glm::value_ptr(material.ambient));
      glUniform4fv(context->materialDiffuseLocation, 1, glm::value_ptr(material.diffuse));
      glUniform4fv(context->materialSpecularLocation, 1, glm::value_ptr(material.specular));
      glUniform1fv(context->materialShininessLocation, 1, &material.shininess);
    }

    // modelMatrix
    glUniformMatrix4fv(context->modelLocation, 1, GL_FALSE, glm::value_ptr(modelMatrix));

    glDrawElements(GL_TRIANGLES, numIndices, GL_UNSIGNED_INT, 0);
  }

  void setPosition(float x, float y, float z) {
    pos = glm::vec3(x, y, z);
  }
  void setSize(float x, float y, float z) {
    size = glm::vec3(x, y, z);
  }

  void update() {
    modelMatrix = glm::mat4(1.0f);
    modelMatrix = glm::translate(modelMatrix, pos);
    modelMatrix = glm::scale(modelMatrix, size);
  }

  glm::mat4 get_m() {
    return modelMatrix;
  }

}; // class Shape



class Light
{
protected:
  const Context *context;

  glm::uint materialEnable;
  GLuint vao[1]; // vertices array buffer
  Model* shape;
  glm::uint numIndices;
  glm::uint numVertices;

  Material  white;
  Material  material;

  glm::mat4 modelMatrix;
  glm::vec3 pos;
  glm::vec3 size;

public:
  // constructor
  Light(const Context* c) : context(c) {
    numIndices = 0;
    numVertices = 0;
    pos = glm::vec3(0.0f,0.0f,0.0f);
    size = glm::vec3(1.0f,1.0f,1.0f);
    materialEnable = false;

    Material tmp = {
      glm::vec4(0.8f, 0.8f, 0.8f, 1.0f),
      glm::vec4(0.2f, 0.2f, 0.2f, 1.0f),
      glm::vec4(1.0f, 1.0f, 1.0f, 1.0f),
      1.0
    };
    white = tmp;
  }

  void setup(const Model* shape)
  {
    GLuint offset;
    GLuint vbo[2];

    if (!shape) {
      std::cerr << "shape is NULL" << "\n";
    }

    numIndices = shape->numIndices;
    numVertices = shape->numVertices;

    // Create VBO
    glGenBuffers(2, vbo);

    // Create and bind VAO
    glGenVertexArrays(1, vao);
    glBindVertexArray(vao[0]);
    {
      // --- VBO 0 ---
      glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      glBufferData(GL_ARRAY_BUFFER, numVertices*sizeof(Vertex), shape->vertices, GL_STATIC_DRAW);

      offset = 0;
      glVertexAttribPointer(context->positionLocation, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const void*)offset); 
      offset += sizeof(glm::vec3);
      glVertexAttribPointer(context->normalLocation, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const void*)offset); 

      // Enable Attribute
      glEnableVertexAttribArray(context->positionLocation);
      glEnableVertexAttribArray(context->normalLocation);

      // --- VBO 1 ---
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo[1]);
      glBufferData(GL_ELEMENT_ARRAY_BUFFER, numIndices*sizeof(glm::uint), shape->indices, GL_STATIC_DRAW);

    }
    glBindVertexArray(0);

    glDeleteBuffers(2, vbo);

    // material setting
    if (shape->materials.size() > 0) {
      material = shape->materials[0];
      materialEnable = true;
    }
  }

  void render()
  {
    // VAO - vertex, normal
    glBindVertexArray(vao[0]);

    glUniform3fv(context->lightPositionLocation, 1, &pos[0]);
    if (materialEnable == 1) {
      glUniform4fv(context->lightAmbientLocation, 1, glm::value_ptr(material.ambient));
      glUniform4fv(context->lightDiffuseLocation, 1, glm::value_ptr(material.diffuse));
      glUniform4fv(context->lightSpecularLocation, 1, glm::value_ptr(material.specular));
    }
    glUniform4fv(context->materialAmbientLocation, 1, glm::value_ptr(white.ambient));
    glUniform4fv(context->materialDiffuseLocation, 1, glm::value_ptr(white.diffuse));
    glUniform4fv(context->materialSpecularLocation, 1, glm::value_ptr(white.specular));
    glUniform1fv(context->materialShininessLocation, 1, &white.shininess);

    // Model Matrix
    glUniformMatrix4fv(context->modelLocation, 1, GL_FALSE, glm::value_ptr(modelMatrix));

    glDrawElements(GL_TRIANGLES, numIndices, GL_UNSIGNED_INT, 0);
  }

  void setPosition(float x, float y, float z) {
    pos = glm::vec3(x, y, z);
  }
  void setSize(float x, float y, float z) {
    size = glm::vec3(x, y, z);
  }

  void update() {
    modelMatrix = glm::mat4(1.0f);
    modelMatrix = glm::translate(modelMatrix, pos);
    modelMatrix = glm::scale(modelMatrix, size);
  }

  glm::mat4 get_m() {
    return modelMatrix;
  }

}; // class Light



#endif // __SHAPE_H_
